//---------------------------------------------------------------------------
//  (c)Alex A. Trubetskoy        alextr438@yandex.ru
//---------------------------------------------------------------------------
#include <vcl.h>
#include <Filectrl.hpp>
#pragma hdrstop
#include "Main.h"
#include "About.h"
#include "Properties.h"
#include "BoxesManage.h"
#include "FormEditMessageUnit.h"
#include "FormMoveCopyMessUnit.h"
#include "const.h"                  //constants
#include "SpMessage.h"
#include "GetMessThread.h"
#include "FormSplashUnit.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "SHDocVw_OCX"
#pragma resource "*.dfm"
TMainForm *MainForm;
//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
        : TForm(Owner), pm(new TBoxesManager())
{
   cppWebBrowser->Anchors << akLeft << akTop << akRight << akBottom;
   OleInitialize(NULL);
   pathApp = GetCurrentDir();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::miExitClick(TObject *Sender)
{
   Close();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::miAboutClick(TObject *Sender)
{
   Application->CreateForm(__classid(TformAbout), &formAbout);
   formAbout->ShowModal();        
}
//---------------------------------------------------------------------------
// the user creates a new box
//---------------------------------------------------------------------------
void __fastcall TMainForm::miCreateBoxClick(TObject *Sender)
{
begin:
   Application->CreateForm(__classid(TFormProperties), &FormProperties);
   FormProperties->edName->Text = "Box" + AnsiString(pm->countBoxes + 1);
   if(FormProperties->ShowModal() == mrOk) {
      //if one of fields is empty
      if(!FormProperties->edEmail->Text.Length())
         {ShowMessage("Please introduce e-mail address"); goto begin;}
      if(!FormProperties->edPassw->Text.Length())
         {ShowMessage("Please introduce password"); goto begin;}
      if(!FormProperties->edPOP->Text.Length())
         {ShowMessage("Please introduce POP-server"); goto begin;}
      if(!FormProperties->edSMTP->Text.Length())
         {ShowMessage("Please introduce SMTP-server"); goto begin;}
      if(!FormProperties->edPath->Text.Length())
         {ShowMessage("Please introduce path to the box folder"); goto begin;}
      SetCurrentDir(pathApp);

      //all fields are filled
      if(!pm->createBox(FormProperties))
         return;    //error message was already shown from the boxmanager module

      addBox2TreeView(FormProperties->edName->Text);
   }
}
//---------------------------------------------------------------------------
__fastcall TMainForm::~TMainForm()
{
   if(pm)
      delete pm;
   for(unsigned i = 0; i < mess.size(); ++i)
      if(mess[i])
         delete mess[i];
}
//---------------------------------------------------------------------------
// add the box to TreeView
//---------------------------------------------------------------------------
void __fastcall TMainForm::addBox2TreeView(AnsiString name) {
   int numItem = (pm->countBoxes - 1) + (pm->countBoxes -1) * COUNT_FOLDERS;

   treeView->Items->Add(NULL, name);
   treeView->Items->Item[numItem]->ImageIndex = 0;
   treeView->Items->AddChild(treeView->Items->Item[numItem], "Inbox");
   treeView->Items->AddChild(treeView->Items->Item[numItem], "Outbox");
   treeView->Items->AddChild(treeView->Items->Item[numItem], "Sent");
   treeView->Items->AddChild(treeView->Items->Item[numItem], "Trash");
   treeView->Items->AddChild(treeView->Items->Item[numItem], "Drafts");
}
//---------------------------------------------------------------------------
// read the configuration of boxes from the file when the form is created
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormCreate(TObject *Sender)
{
   Application->CreateForm(__classid(TformSplash), &formSplash);
   formSplash->Show();
   formSplash->Update();

   //set captions for StringGrid columns
   sgMessages->Cells[0][0] = "from";
   sgMessages->Cells[1][0] = "to";
   sgMessages->Cells[2][0] = "subject";
   sgMessages->Cells[3][0] = "created";

   if(FileExists("config.spcf")) {
      int hFile, lenFile, countBytes, i = 0;
      AnsiString list[COUNT_FEATURES];
      AnsiString temp;
      if((hFile = FileOpen("config.spcf", fmOpenRead)) == -1) {
         MessageBox(NULL, "Internal error: unable to open config file  ",
                    "Error", MB_ICONSTOP);
         return;
      }
      lenFile = FileSeek(hFile, 0, 2);
      FileSeek(hFile, 0, 0);
      char* pszBuffer = new char[lenFile + 1];
      countBytes = FileRead(hFile, pszBuffer, lenFile);
      FileClose(hFile);
      while(i < countBytes && pszBuffer[i] == '\n') ++i;
      while(i < countBytes) {
         for(int j = 0; j < COUNT_FEATURES; ++j) {
            //get string
            while(i < countBytes && pszBuffer[i] != '\n')
               list[j] += pszBuffer[i++];
            while(pszBuffer[i] == '\n') ++i;
         }
         //add new box
         bool d = (list[DEL] == "y")? true : false;
         if(!pm->restoreBox(list, d)) {
            MessageBox(NULL, "Internal error  ", "Error", MB_ICONSTOP);
            return;
         }
         addBox2TreeView(list[NAME]);
         for(int j = 0; j < COUNT_FEATURES; ++j)
            list[j] = "";

         while(pszBuffer[i] == '\n') ++i;
      }
      delete[] pszBuffer;
      treeView->FullExpand();
   }
   this->WindowState = wsMaximized;

   Sleep(1000);
   formSplash->Release();
   Show();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::miWriteMessClick(TObject *Sender)
{
   FormEditMessage->path = pm->getPathOfActiveBox();
   FormEditMessage->Show();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::clearGrid()
{
   for(int row = 1; row < sgMessages->RowCount; ++row)
      for(int col = 0; col < (COUNT_MESS_FIELDS - 1); ++col)
         sgMessages->Cells[col][row] = "";
}
//---------------------------------------------------------------------------
// display text of the selected message in memo control
//---------------------------------------------------------------------------
void __fastcall TMainForm::sgMessagesSelectCell(TObject *Sender, int ACol,
      int ARow, bool &CanSelect)
{
   bool html = false;
   int hFile;
   AnsiString path, folder;
   clearWebBrowser();
   //show selected message
   if((unsigned)(ARow - 1) < mess.size()) {
      miMessDelete->Enabled = true;
      miMessCopy->Enabled = true;
      miMessMove->Enabled = true;

      //check if the message is html document
      AnsiString lower = mess[ARow - 1]->text.LowerCase();
      if(lower.Pos("<html") != 0) html = true;
         
      if(html) {
         folder = treeView->Items->Item[getNumberOfSelFolder()]->Text;
         path = pm->getPathOfActiveBox() + folder +"\\html.tmp";
         if(FileExists(path))
            DeleteFile(path);
         Sleep(300);
         hFile = FileCreate(path);
         FileWrite(hFile, mess[ARow - 1]->text.c_str(), mess[ARow - 1]->text.Length());
         FileClose(hFile);
         cppWebBrowser->Navigate(WideString(path));
      }
      else {
         AnsiString temp = mess[ARow - 1]->text;
         for(int i = 1; i <= temp.Length(); ++i) {
            switch(temp[i]) {
               case '\n':
                  temp.Delete(i, 1);
                  temp.Insert("<br>", i);  i += 3; break;
               case '<' :
                  temp.Delete(i, 1);
                  temp.Insert("&lt;", i);  i += 3; break;
               case '>' :
                  temp.Delete(i, 1);
                  temp.Insert("&gt;", i);  i += 3; break;
               case '&' :
                  temp.Delete(i, 1);
                  temp.Insert("&amp;", i); i += 4; break;
            }
         }
         temp.Insert("<html><body>", 1);
         temp += "</body></html>";
         folder = treeView->Items->Item[getNumberOfSelFolder()]->Text;
         path = pm->getPathOfActiveBox() + folder +"\\html1.tmp";
         if(FileExists(path))
            DeleteFile(path);
         Sleep(300);
         hFile = FileCreate(path);
         FileWrite(hFile, temp.c_str(), temp.Length());
         FileClose(hFile);
         cppWebBrowser->Navigate(WideString(path));
      }
   }
   else {
      clearWebBrowser();
      miMessDelete->Enabled = false;
      miMessCopy->Enabled = false;
      miMessMove->Enabled = false;
   }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::actionTreeClickExecute(TObject *Sender)
{
   //clear the storage for messages
   for(unsigned i = 0; i < mess.size(); ++i)
      if(mess[i])
         delete mess[i];
   mess.clear();
   //clear the grid & memo
   clearGrid();
   clearWebBrowser();
   TGridRect curRect = sgMessages->Selection;
   int numOfMess = curRect.Top;
   if((unsigned)numOfMess > mess.size()) {
      miMessDelete->Enabled = false;
      miMessCopy->Enabled = false;
      miMessMove->Enabled = false;
   }
   
   //set active box
   TTreeNode* sel = treeView->Selected;
   int num = -1;
   if(sel != NULL) {
      num = getNumberOfSelFolder();
      if(num == -1) {
         MessageBox(NULL, "Internal error: unable to set active box  ",
                    "Error", MB_ICONSTOP);
      }
      else
         pm->setActiveBox(num / (COUNT_FOLDERS + 1));
   }

   //retrieve messages
   if((num % (COUNT_FOLDERS + 1)) != 0) {
      AnsiString path = pm->getPathOfActiveBox(), end = "", folder;
      AnsiString list[COUNT_MESS_FIELDS];
      int hFile, lenFile, countBytes, i = 0, numMess = 1;

      folder = treeView->Items->Item[num]->Text;

      if(!DirectoryExists(path)) {
         AnsiString mess = "Directory " + path + " doesn't exist!"
                        + " Unable open the box. ";
         MessageBox(NULL, mess.c_str(), "Error", MB_ICONSTOP);
         return;
      }

      path += (folder + "\\message.spmg");
      if(FileExists(path)) {
         if((hFile = FileOpen(path, fmOpenRead)) == -1) {
            MessageBox(NULL, "Unable to open message file  ", "Error", MB_ICONSTOP);
            return;
         }
         lenFile = FileSeek(hFile, 0, 2);
         char* pszBuffer = new char[lenFile + 1];
         FileSeek(hFile, 0, 0);
         countBytes = FileRead(hFile, pszBuffer, lenFile);
         FileClose(hFile);
         while(i < countBytes && pszBuffer[i] == '\n') ++i;

         while(i < countBytes) {
            //ignore "beginofmessagesp" string
            while(i < countBytes && pszBuffer[i] != '\n') ++i;
            while(i < countBytes && pszBuffer[i] == '\n') ++i;
            for(int j = 0; j < COUNT_MESS_FIELDS - 1; ++j) {
               //get string
               while(i < countBytes && pszBuffer[i] != '\n')
                  list[j] += pszBuffer[i++];
               while(pszBuffer[i] == '\n') ++i;
            }
            while(i < countBytes && pszBuffer[i] != (char)0x01)
               list[MESS_TEXT] +=  pszBuffer[i++];
            end = "";
            while(i < countBytes && pszBuffer[i] != '\n')
               end += pszBuffer[i++];
            if(strstr(end.c_str(), "endofmessagesp") == NULL) {
               MessageBox(NULL, "The message file is corrupted  ", "Error", MB_ICONSTOP);
               return;
            }
            while(i < countBytes && pszBuffer[i] != '\n') ++i;

            //show the message in the StringGrid
            for(int col = 0;
                col < (COUNT_MESS_FIELDS - 2) && col < sgMessages->ColCount;
                ++col)
            {
               sgMessages->Cells[col][numMess] = list[col + 1];
            }
            mess.push_back(new TSpMessage(list));
            ++numMess;
            if((numMess + 1) > sgMessages->RowCount)
               ++sgMessages->RowCount;

            for(int j = 0; j < COUNT_MESS_FIELDS; ++j)
               list[j] = "";
            while(pszBuffer[i] == '\n') ++i;
         }
         delete[] pszBuffer;
      }
      else {          //there is no message file
         clearGrid();
      }
   }
   else {
      clearGrid();
   }
}
//---------------------------------------------------------------------------
// get number of the selected box
//---------------------------------------------------------------------------
int __fastcall TMainForm::getNumberOfSelFolder()
{
   int num = -1;
   for(int i = 0; i < treeView->Items->Count; ++i) {
      if(treeView->Items->Item[i]->Selected == true) {
         num = i;
         break;
      }
   }
   return num;
}
//---------------------------------------------------------------------------
// the user gets new messages for the box
//---------------------------------------------------------------------------
void __fastcall TMainForm::actionGetMessExecute(TObject *Sender)
{
   TGetMessThread* pThread = new TGetMessThread(true);
   pThread->FreeOnTerminate = true;
   pThread->setTask(true);
   pThread->Resume();
}
//---------------------------------------------------------------------------
// the user send messages from the box
//---------------------------------------------------------------------------
void __fastcall TMainForm::actionSendMessExecute(TObject *Sender)
{
   TGetMessThread* pThread = new TGetMessThread(true);
   pThread->FreeOnTerminate = true;
   pThread->setTask(false);
   pThread->Resume();
}
//---------------------------------------------------------------------------
// delete selected message
//---------------------------------------------------------------------------
void __fastcall TMainForm::actionDelMessExecute(TObject *Sender) {
   //delete the message from the file
   updateMessFile(1);
   //update StringGrid contens
   //& delete the message from mess vector
   actionTreeClickExecute(Sender);
}
//---------------------------------------------------------------------------
// move selected message
//---------------------------------------------------------------------------
void __fastcall TMainForm::actionMoveMessExecute(TObject *Sender)
{
   copyMessage(true);
   //update StringGrid contens
   //& delete the message from mess vector
   actionTreeClickExecute(Sender);
}
//---------------------------------------------------------------------------
// copy selected message
//---------------------------------------------------------------------------
void __fastcall TMainForm::actionCopyMessExecute(TObject *Sender)
{
   copyMessage(false);
}
//---------------------------------------------------------------------------
// update message file when the user deletes, copies or moves any message
//---------------------------------------------------------------------------
void __fastcall TMainForm::updateMessFile(short op, AnsiString* pStr)
{
   //op: 1-delete, 2-move, 3-copy
   //get selected message
   TGridRect curRect = sgMessages->Selection;
   int numOfMess = curRect.Top;
   if((unsigned)numOfMess > mess.size())
      return;

   int num = getNumberOfSelFolder();
   AnsiString folder = treeView->Items->Item[num]->Text;
   AnsiString path = pm->getPathOfActiveBox() + folder +"\\message.spmg";
   //delete the whole message file if there's only 1 message in the folder
   if((numOfMess == 1) && (mess.size() == 1)) {
      if(FileExists(path))
         DeleteFile(path);
   }
   else
   if((unsigned)numOfMess <= mess.size()) {
      //delete the message from file and mess vector
      int hFile, hFile2, lenFile;
      if(num == -1) {
         MessageBox(NULL, "Internal error: unable to get active box  ",
                    "Error", MB_ICONSTOP);
         return;
      }
      if(FileExists(path)) {
         if((hFile = FileOpen(path, fmOpenRead)) == -1) {
            MessageBox(NULL, "Unable to open message file  ", "Error", MB_ICONSTOP);
            return;
         }
         lenFile = FileSeek(hFile, 0, 2);
         FileSeek(hFile, 0, 0);
         char* pszBuffer = new char[lenFile + 1];
         AnsiString bufferTemp, newText;
         FileRead(hFile, pszBuffer, lenFile);
         FileClose(hFile);

         //search bounds of the selected message
         int i = 0, j, index_begin, index_end;
         while(numOfMess > 0 && i < lenFile) {
            while(pszBuffer[i] == '\n') ++i;
            for(; (pszBuffer[i] != '\n' && i < lenFile); ++i) {
               bufferTemp += pszBuffer[i];
            }
            if(strstr(bufferTemp.c_str(), "eginofmessagesp")) {
               --numOfMess;
            }
            bufferTemp = "";
         }
         j = i - 1;
         while(pszBuffer[j] != '\n') --j;
         index_begin = j;
         while(i < lenFile) {
            while(pszBuffer[i] == '\n') ++i;
            for(; (pszBuffer[i] != '\n' && i < lenFile); ++i) {
               bufferTemp += pszBuffer[i];
            }
            if(strstr(bufferTemp.c_str(), "ndofmessagesp")) {
               break;
            }
            bufferTemp = "";
         }
         index_end = i;

         //store the selected message
         if((op == 2 || op == 3) && pStr) {  //copy or move
            for(i = index_begin; i <= index_end; ++i) {
               (*pStr) += pszBuffer[i];
            }
         }

         //update the source message file
         if(op == 1 || op == 2) {    //move or delete
            RenameFile(path, path + ".tmp");
            if((hFile2 = FileCreate(path)) == -1) {
               MessageBox(NULL, "Unable to create new message file  ", "Error", MB_ICONSTOP);
               RenameFile(path + ".tmp", path);    //undo
               if(pszBuffer)delete pszBuffer;
               return;
            }
            for(i = 0; i < lenFile; ++i) {
               if(i > index_begin && i < index_end)
                  continue;
               newText += pszBuffer[i];
            }
            if(FileWrite(hFile2, newText.c_str(), newText.Length()) == -1) {
               MessageBox(NULL, "Unable to write data to message file  ", "Error", MB_ICONSTOP);
               FileClose(hFile2);
               DeleteFile(path);
               RenameFile(path + ".tmp", path);    //undo
               if(pszBuffer)delete pszBuffer;
               return;
            }
            FileClose(hFile2);
            DeleteFile(path + ".tmp");
         }
         if(pszBuffer)delete pszBuffer;
      }
   }
}
//---------------------------------------------------------------------------
// the function is used when a message must be copied or moved
//---------------------------------------------------------------------------
void __fastcall TMainForm::copyMessage(bool move)
//if move - move message, if !move - copy message
{
   //get selected message
   TGridRect curRect = sgMessages->Selection;
   AnsiString path, folder, message;
   int numOfMess = curRect.Top, hFile;
   if((unsigned)numOfMess > mess.size())  //no message was selected
      return;

   //show the form allowing the user to select the destination folder
   AnsiString fileName = "treeview.tmp";
   treeView->SaveToFile(fileName);
   if(!move)
      FormMoveCopyMess->Caption = "Copy Message To";
   else
      FormMoveCopyMess->Caption = "Move Message To";
      
   FormMoveCopyMess->paintTreeView(fileName);
   int mRes = FormMoveCopyMess->ShowModal();
   if(mRes == mrCancel)
      return;
   else
   if(mRes == mrOk) {
      int num = FormMoveCopyMess->getNumberOfSelFolder();
      if(num == -1)
         return;
      path = pm->getPathOfBox(num / (COUNT_FOLDERS + 1));
      folder = FormMoveCopyMess->getPath();
      if(folder == "-1")
         return;
      path += folder;

      if(FileExists(path)) {
         if((hFile = FileOpen(path, fmOpenWrite)) == -1) {
            MessageBox(NULL, "Unable to open destination message file  ", "Error", MB_ICONSTOP);
            return;
         }
      }
      else {
         if((hFile = FileCreate(path)) == -1) {
            MessageBox(NULL, "Unable to create new message file  ", "Error", MB_ICONSTOP);
            return;
         }
      }

      if(!move)
         updateMessFile(3, &message);    //copy
      else
         updateMessFile(2, &message);    //move
      message += "\n\n\n";
      FileSeek(hFile, 0, 2);
      FileWrite(hFile, message.c_str(), message.Length());
      FileClose(hFile);
   }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::sgMessagesMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
   if(Button == mbRight) {
      pmForSG->Popup(Left + sgMessages->Left + X, Top + sgMessages->Top + Y);
   }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::clearWebBrowser()
{
   int hFile;
   if(!FileExists("clear.tmp")) {
      if((hFile = FileCreate("clear.tmp")) == -1) {
         MessageBox(NULL, "Unable to create temporary file  ",
                    "Error", MB_ICONSTOP);
         return;
      }
      AnsiString buf = "<html></html>";
      if(FileWrite(hFile, buf.c_str(), buf.Length()) == -1) {
         MessageBox(NULL, "Internal error  ", "Error", MB_ICONSTOP);
         FileClose(hFile);
         return;
      }
   }
   Sleep(300);
   cppWebBrowser->Navigate(WideString(pathApp + "\\clear.tmp"));
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormClose(TObject *Sender, TCloseAction &Action)
{
    Application->Terminate();    
}
//---------------------------------------------------------------------------

